local mod_GenStream = require "mod_GeneralStream"

local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

require "kmCore"

local _M = {
}

---------------------------------------------

local function FormatURL( url, user, pass )
	local protocol, orig_url_tail
	local user_in_url, pass_in_url

	local s

	protocol, s = url:match( "(%w+)://(.+)$" )
	if not protocol then
		return url:match("(%w+):.*"), url
	end

	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
		local s1 = s:match("(.-)/(.+)")
		if s1 then
			user_in_url = s1:match( "([^/]-)@./+$" )
		end
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if user and user ~= "" then
		user_in_url = user
		pass_in_url = pass
	end

	if user_in_url and user_in_url ~= "" then
		url = protocol .. "://" .. user_in_url
		if pass_in_url then
			url = url .. ":" .. pass_in_url
		end
		url = url .. "@" .. orig_url_tail
	else
		url = protocol .. "://" .. orig_url_tail
	end
	return protocol, url
end

local stream_meta = {}

function stream_meta:SpecialInit()
	--Step2: Create network media source
	self.networkSource = KM_STREAMING.networkSource.createNew()
	if not self.networkSource then
		_dbg.Error( "NET_STREAM", self.id, ": Fail to create network media puller" )
		self:Destroy()
		return false
	end

	self.syncSource:BindMediaSource( self.networkSource:GetMediaSource(), true ) --Since previous is just only network source, I can reuse the crossbuffer
	--TODO: Start seek the network stream?
	return true
end

function stream_meta:SpecialDestroy()
	self.networkSource = nil
end

function stream_meta:CheckURLAndOptions(url, options)
	options = options or {}
	local stream_options = self.stream_options or {}
	local url_type, url = FormatURL(url, options.username, options.password) 
	if not url_type then
		_dbg.Error( 'NET_STREAM', self.id, "Invalid URL type of:", url )
		return false, nil
	end

	--TODO:
	stream_options.read_timeout=5000000
	stream_options.connect_timeout=5000000
	stream_options.connIntv=3000000

	local updated = false
	if self.type ~= url_type then
		self.type = url_type
		updated = true
	end

	if url_type == "rtsp" then
		local tp = options.transport or "tcp"
		if stream_options.rtsp_transport ~= tp then
			stream_options.rtsp_transport = tp
			updated = true
		end
	elseif url_type == "rtmp" then
		stream_options.rtsp_transport = nil
	elseif url_type == "udp" then
		stream_options.rtsp_transport = nil
	elseif url_type == "rtp" then
		stream_options.rtsp_transport = nil
		--TODO: Generate the SDP.
	elseif url_type == "http" then
		stream_options.rtsp_transport = nil
	else
		stream_options.rtsp_transport = nil
	end

	self.stream_options = stream_options

	return updated, url
end

function stream_meta:ChangeSource(url,options)
	local updated
	url = url or self.url or ""
	if options then
		updated,url = self:CheckURLAndOptions(url,options)
        if not url then
			return
		end
	end

	if self.url ~= url or updated then
		self.url = url
		if not self:CheckStartActions() then
			return
		end
        self:Stop(nil, true) --Stop all but don't free decoder.
		self:Start()
	end
end

function stream_meta:SpecialStart(group, action)
	self.networkSource:Open( self.url, self.stream_options )
	_dbg.Info( "NET_STREAM", self.id, ": Created stream puller for url:", self.url )
	return true
end

function stream_meta:SpecialStop(group, action)
	self.networkSource:Close()
	_dbg.Info( "NET_STREAM", self.id, ": Destroied stream puller" )
	return true
end

function stream_meta:GetStreamStatus(group)
	local with_act, acts = self:CheckStartActions()
	local stream_start = false
	if with_act then
		local k,v
		for k,v in pairs(acts) do
			stream_start = true
			break
		end
	end

	local ret = {
		stream_start = stream_start
	}

	if not self.networkSource then
		ret.live = "closed"
		ret.resolution = "?x?"
		ret.bitrate = 0
		ret.video = "NONE"
		ret.audio = "NONE"
		ret.frameRate = 0
	else
		local status = self.networkSource:GetStatus()
		-- local pl = require "pl.pretty"
		-- pl.dump(status)
		if not status then
			ret.live = "error"
			ret.resolution = "?x?"
			ret.bitrate = 0
			ret.video = "NONE"
			ret.audio = "NONE"
			ret.frameRate = 0
		else
			ret.live = status.live and status.live.status or "error"
			if status.video then
				ret.resolution = '' .. (status.video.width or "0") .. "x" .. (status.video.height or "0") .. "@" .. (status.video.frameRate or "0") .. "Hz"
				ret.bitrate = status.video.bitrate or 0
				ret.video = status.video.codec or "H264"
				ret.frameRate = status.video.frameRate or 0
			else
				ret.video = "NONE"
				ret.resolution = "?x?"
				ret.bitrate = 0
				ret.frameRate = 0
			end

			if status.audio then
				ret.audio = status.audio.codec or "AAC"
				ret.bitrate = ret.bitrate + (status.audio.bitrate or 0)
			else
				ret.audio = "NONE"
			end
		end
	end
	return ret
end

function stream_meta:GetMediaSource()
	if self.syncSource then
		return self.syncSource:GetMediaSource()
	-- if self.networkSource then
	-- 	return self.networkSource:GetMediaSource()
	else
		return nil
	end
end

stream_meta = mod_GenStream.ConstructStreamMeta( stream_meta )

----------------------------------

function _M.Init()
end

function _M.Create(id, buffer_option, islocal)
	local stream = mod_GenStream.CheckStreamExist(id)
	if stream then
		_dbg.Warn( "NET_STREAM", "The creating stream '" .. id .. "' is already exist." )
		return stream
	end

	if not buffer_option or buffer_option == "" then
		buffer_option = "live:200:200:500:0:0"
	end

	if not islocal  then
		islocal = false
	end

	stream = {
		id = id,
		type = "unknown",
		start_actions = {},
		audio_outputs = {},
		audio_enabled = {},
		vpss_using = {},
		current_audio_intfs = {},
		stream_openned = false,
		buffer_option = buffer_option,
		url = nil,
		stream_options = nil,
		islocal = islocal
	}

	setmetatable( stream, stream_meta )
	stream:Init()
	return stream
end

return _M
