local mod_GenStream = require 'mod_GeneralStream'

local coaf = require 'coaf.core'
local _dbg = require 'coaf.debug'

require 'kmCore'

local _M = {}

local stream_meta = {}

function stream_meta:deepcompare(t1,t2,ignore_mt)
    local ty1 = type(t1)
    local ty2 = type(t2)
    if ty1 ~= ty2 then return false end
    -- non-table types can be directly compared
    if ty1 ~= 'table' and ty2 ~= 'table' then return t1 == t2 end
    -- as well as tables which have the metamethod __eq
    local mt = getmetatable(t1)
    if not ignore_mt and mt and mt.__eq then return t1 == t2 end
    for k1,v1 in pairs(t1) do
    local v2 = t2[k1]
    if v2 == nil or not self:deepcompare(v1,v2) then return false end
    end
    for k2,v2 in pairs(t2) do
    local v1 = t1[k2]
    if v1 == nil or not self:deepcompare(v1,v2) then return false end
    end
    return true
end

function stream_meta:SpecialInit()
    self.srtrecv = KM_STREAMING.srtService.createNew()
    if not self.srtrecv then
        _dbg.Error('SRT_RECV', self.id, ' fail to create srt recv service')
        self.Destroy()
        return false
    end

    self.syncSource:BindMediaSource(self.srtrecv:GetMediaSource(), true)
    return true
end

function stream_meta:SpecialDestroy()
    self.srtrecv = nil
    _dbg.Info('SRT_RECV', self.id, ': destroy SRT RECV service')
    return true
end

function stream_meta:GetAesKeyLen(aesName)
    local aesKeyLen
    if aesName == 'AES-128' or aesName == 'aes-128' then
        aesKeyLen = 16
    elseif aesName == 'AES-192' or aesName == 'aes-192' then
        aesKeyLen = 24
    elseif aesName == 'AES-256' or aesName == 'aes-256' then
        aesKeyLen = 32
    else
        aesKeyLen = 0
    end
    return aesKeyLen
end


function stream_meta:CheckURLAndOptions(url, options)
    local stream_options = {}
    local key_lenth = #options.passphrase

    if options.encryption ~= 'none' then
        if key_lenth == 0 then
            print('*** WARNING: Invalid SRT PARA encryption_key/key_len, disabled. [key-Val is NULL]')
            options.encryption = 'none'
        elseif key_lenth < 10 then
            print('*** ERROR: Invalid SRT PARA srt_encryption_key disabled. [Key-Val length less than 16]')
            return false, nil
        end
	end
	
	local protocol, s = url:match( "(%w+)://(.+)$" )
	if not s then
		print('*** ERROR: Invalid SRT PARA. URL=',url)
		return false, nil
	end

	local addr, port = s:match( "(.*):(%d+)$" )
	if not port then
		print('*** ERROR: Invalid SRT PARA. URL=',url)
		return false, nil
	end

    local updated = false
    --TODO:
    stream_options.srt_connect_mode = options.connectionMode
    stream_options.srt_encryption_key_len = self:GetAesKeyLen(options.encryption)
    stream_options.srt_encryption_key = options.passphrase
    stream_options.srt_bandwidth_overhead = options.bandwidth * 0.01
    stream_options.srt_latency = options.latency
	stream_options.srt_mtu = options.payloadSize
	stream_options.showAdvanced = options.showAdvanced
	--stream_options.listenerPort = options.listenerPort
	stream_options.listenerPort = port
    if not addr then
        stream_options.address = ''
    else
        stream_options.address = addr
    end
    if options.connectionMode:lower() == 'listener' then
        stream_options.address = ''
    end

	-- srt stream info analyse speed
	local con_speed = options.connect_speed or 2
	if stream_options.connect_speed ~= con_speed then
		stream_options.connect_speed = con_speed
	end

	if con_speed == 0 then
		stream_options.analyzeduration = 1500000   --1.5s
		stream_options.probesize 	   = 1500000   --1.5m
	elseif con_speed == 1 then
		stream_options.analyzeduration = 3000000
		stream_options.probesize 	   = 3000000
	elseif con_speed == 3 then
		stream_options.analyzeduration = 8000000
		stream_options.probesize	   = 8000000
	else
			--nothing todo
	end
	stream_options.srt_streamid = options.srt_streamid

	local ret = self:deepcompare(self.stream_options, stream_options, nil)
	if not ret then 
		self.stream_options = stream_options
		updated = true
	end
    return updated, url
end

function stream_meta:ChangeSource(url, options)
    local updated
    url = url or self.url or ''
    if options then
        updated, url = self:CheckURLAndOptions(url, options)
        if not url then
            return
        end
    end

    if self.url ~= url or updated then
        self.url = url
        if not self:CheckStartActions() then
            return
        end
        self:Stop(nil, true) --Stop all but don't free decoder.
        self:Start()
    end
end

function stream_meta:SpecialStart(group, action)
    self.srtrecv:set_options(self.stream_options.address, self.stream_options.listenerPort, self.stream_options)
    self.srtrecv:start()
    _dbg.Info('SRT_RECV', self.id, ': started SRT RECV service,  port is :', self.url)
end

function stream_meta:SpecialStop(group, action)
    self.srtrecv:stop()
    _dbg.Info('SRT_RECV', self.id, ': stop SRT RECV service')
    return true
end

function stream_meta:GetStreamStatus(group)
    local with_act, acts = self:CheckStartActions()
    local output_start = false
    local stream_start = false
    if with_act then
        local k, v
        for k, v in pairs(acts) do
            if k == 'output' then
                output_start = true
                stream_start = true
                break
            else
                stream_start = true
            end
        end
    end

    local ret = {
        output_start = output_start,
        stream_start = stream_start
    }

    if not self.srtrecv then
        ret.live = 'closed'
        ret.resolution = '?x?'
        ret.bitrate = 0
        ret.video = 'NONE'
        ret.audio = 'NONE'
        ret.frameRate = 0
    else
        local status = self.srtrecv:GetStatus()
        if not status then
            ret.live = 'error'
            ret.resolution = '?x?'
            ret.bitrate = 0
            ret.video = 'NONE'
            ret.audio = 'NONE'
            ret.frameRate = 0
        else
            ret.live = status.live and status.live.status or 'error'
            if status.video then
                ret.resolution =
                    '' ..
                    (status.video.width or '0') ..
                        'x' .. (status.video.height or '0') .. '@' .. (status.video.frameRate or '0') .. 'Hz'
                ret.bitrate = status.video.bitrate or 0
                ret.video = status.video.codec or 'H264'
                ret.frameRate = status.video.frameRate or 0
            else
                ret.video = 'NONE'
                ret.resolution = '?x?'
                ret.bitrate = 0
                ret.frameRate = 0
            end

            if status.audio then
                ret.audio = status.audio.codec or 'AAC'
                ret.bitrate = ret.bitrate + (status.audio.bitrate or 0)
            else
                ret.audio = 'NONE'
            end
        end
    end
    return ret
end

function stream_meta:GetMediaSource()
    --if self.syncSource then
    --      return self.syncSource:GetMediaSource()
    if self.srtrecv then
        return self.srtrecv:GetMediaSource()
    else
        return nil
    end
end

function stream_meta:GetSrtStatus(group)
    local ret = {}
    if not self.srtrecv then
        ret.srtMode = 'Unkonwn'
        ret.liveStatus = 'Disconnected'
        ret.pktRcvRetrans = 0
        ret.pktRcvLoss = 0
        ret.pktRcvLossTotal = 0
        ret.byteRecvTotal = 0
        ret.rtt = 0
        ret.bandwidth = 0
    else
        local srtStatus = self.srtrecv:GetSrtStatus()
        if not srtStatus then
            ret.srtMode = 'Unkonwn'
            ret.liveStatus = 'Disconnected'
            ret.pktRcvRetrans = 0
            ret.pktRcvLoss = 0
            ret.pktRcvLossTotal = 0
            ret.byteRecvTotal = 0
            ret.rtt = 0
            ret.bandwidth = 0
        else
            ret.srtMode = srtStatus.mode
            ret.liveStatus = srtStatus.status
            ret.pktRcvRetrans = srtStatus.retrans_packets
            ret.pktRcvLoss = srtStatus.lost_packets
            ret.pktRcvLossTotal = srtStatus.lost_packets_total
            ret.byteRecvTotal = srtStatus.byte_recv_total
            ret.rtt = srtStatus.rtt
            ret.bandwidth = srtStatus.bandwidth
        end
    end
    return ret
end

stream_meta = mod_GenStream.ConstructStreamMeta(stream_meta)

----------------------------------

function _M.Init()
end

function _M.Create(id, buffer_option)
    local stream = mod_GenStream.CheckStreamExist(id)
    if stream then
        _dbg.Warn('NET_STREAM', "The creating stream '" .. id .. "' is already exist.")
        return stream
    end

    if not buffer_option or buffer_option == '' then
        buffer_option = 'live:200:200:500:0:0'
    end

    stream = {
        id = id,
        --type = "unknown",
        type = 'srt',
        decoder_group_id = "min_group",
        decoderGroup = "decoder_group",
        decoder_channel = nil,
        start_actions = {},
        audio_outputs = {},
        audio_enabled = {},
        vpss_using = {},
        current_audio_intfs = {},
        stream_openned = false,
        buffer_option = buffer_option,
        url = nil,
        stream_options = nil,
        origin = "manual"
    }

    setmetatable(stream, stream_meta)
    stream:Init()
    return stream
end

return _M
