local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local cfg_streamer = require "streamer_config"
local streams = require "mod_Streams"

local CO = require "mod_coservice"

local _done, features = pcall(require, "features")
if not _done then
	features = {}
elseif type(features) ~= "table" then
	features = _G
end


require "kmCore"

local MAX_SUPPORT_STREAMS = 6
local MAX_SUPPORT_PUBPOINTS = 9

if features.DEVICE_TYPE == "E_DOCTOR" or features.DEVICE_TYPE == "E_DOCTOR_S" then
	MAX_SUPPORT_STREAMS = 8
	MAX_SUPPORT_PUBPOINTS = 24
end

local _M = {
	pubpoint_id_base = (features.DEVICE_TYPE == "E_DOCTOR" or features.DEVICE_TYPE == "E_DOCTOR_S") and 9 or 1,
	co_service = nil,
	config = nil,
	rtsp_servers = {},
	pub_points = {},
	pub_points_count = 0,
	ndi_url = "",
	full_ndiChannel = nil
}

------------------------------------

local pubpoint_meta = {}

function pubpoint_meta:Init(id,cfg)
	if not id or not cfg then
		_dbg.Error( "Fail to initialize publish point: with invalid id or configure!" )
		return false
	end

	self.id = 1
	self.config = cfg
	self.stream_id_base = 1	
	self.streams = {}
	self.streams_count = 0

	self.expose_path = "/streamer/" .. self.id

	if not _M.co_service then
		_dbg.Warn( "STREAMER", "Publish point of:", id, " cannot expose for co-service because no valid service!" )
	else
		_M.co_service:ExposeObject( self.expose_path, self )
		_dbg.Info( "STREAMER", "Publish point of:", id, " exposed for co-service, path:", self.expose_path )
	end

	if cfg.source_stream and cfg.source_stream ~= ""then
		self.stream_obj = streams.GetStreamObject( cfg.source_stream ,"main")
		self.sub_stream_obj =  streams.GetStreamObject( cfg.source_stream ,"sub")
	else
		self.stream_obj = nil
		self.sub_stream_obj = nil
	end

	if self.stream_obj then
		--Group is "pub:<id>"
		self.stream_obj:Start("pub:" .. self.id, "stream")
	end
	if self.sub_stream_obj then
		self.sub_stream_obj:Start("pub:" .. self.id, "stream")
	end


	--Local stream, no switcher.
	if id:match("local") then
		if self.stream_obj then
			self.mediaSource = self.stream_obj:GetMediaSource()
		else
			self.mediaSource = nil
		end

		if self.sub_stream_obj then
			self.sub_mediaSource = self.sub_stream_obj:GetMediaSource()
		else
			self.sub_mediaSource = nil
		end

	else
		--Step1: Create media switcher
		_dbg.Info( "STREAMER", "Publish point:", id, " media switcher is created." )
		self.mediaSW = KM_STREAMING.SwitchSource.createNew()
		self.sub_mediaSW = KM_STREAMING.SwitchSource.createNew()

		--Step2: After media switcher synchornizer (to correct-up timestamp)
		_dbg.Info( "STREAMER", "Publish point:", id, " create media synchronizer for switcher" )
		self.afterSwSync = KM_STREAMING.SyncSource.createNew()
		self.sub_afterSwSync = KM_STREAMING.SyncSource.createNew()
		self.afterSwSync:SetBypass()	--This is only a bypass using.
		self.sub_afterSwSync:SetBypass()	--This is only a bypass using.

		--TODO: NOTE here:
		--  If (video) codec specified '*', that track must wait selected track be ready.
		--  And audio is specified by default 'AAC' for this application, so, if no audio in
		--  selected media source, it *FAKE* a AAC audio track (but could not get any stream data)
		self.mediaSW:SetTrackDefaultCodec( "video", "*" )
		self.mediaSW:SetTrackDefaultCodec( "audio", "*" )
		self.sub_mediaSW:SetTrackDefaultCodec( "video", "*" )
		self.sub_mediaSW:SetTrackDefaultCodec( "audio", "*" )
		--[[
		self.mediaSW:SetTrackDefaultCodec( "audio", "AAC", {
			sampling = 48000,
			channels = 2,
			sampleSize = 1024,
			sampleWidth = 16,
			aac_format = 0, --TODO:
			aac_profile = 1, --TODO:
			aac_config = "\x11\x90"
		})
		--]]
		if self.stream_obj then
			self.mediaSW:SelectTrack( "video", self.stream_obj:GetMediaSource() )
			self.mediaSW:SelectTrack( "audio", self.stream_obj:GetMediaSource() )
		else
			self.mediaSW:SelectTrack( "video", nil )
			self.mediaSW:SelectTrack( "audio", nil )
		end

		if self.sub_stream_obj then
			self.sub_mediaSW:SelectTrack( "video", self.sub_stream_obj:GetMediaSource() )
			self.sub_mediaSW:SelectTrack( "audio", self.sub_stream_obj:GetMediaSource() )
		else
			self.sub_mediaSW:SelectTrack( "video", nil )
			self.sub_mediaSW:SelectTrack( "audio", nil )
		end

		self.afterSwSync:BindMediaSource( self.mediaSW:GetMediaSource() )
		self.mediaSource = self.afterSwSync:GetMediaSource()

		self.sub_afterSwSync:BindMediaSource( self.sub_mediaSW:GetMediaSource() )
		self.sub_mediaSource = self.sub_afterSwSync:GetMediaSource()


	end

	if not cfg.streams then
		return true
	end

	local sid, scfg
	local max_id
	for sid, scfg in pairs(cfg.streams) do
		if not max_id or tonumber(sid) > max_id then
			max_id = tonumber(sid)
		end
		if self:CreateStream( scfg.type, sid ) then
			self:SetStreamParameters( sid, nil, true )
		end
	end

	self.stream_id_base = max_id and (max_id + 1) or 1

	return true
end

function pubpoint_meta:Destroy()
	local list = {}
	local sid, s
	for sid,s in pairs(self.streams) do
		table.insert( list, sid )
	end

	for s=1, #list do
		self:RemoveStream( list[s], true ) --Don't remove configures while destroying.
	end

	if self.stream_obj then
		--Group is "pub:<id>"
		self.stream_obj:Stop("pub:" .. self.id, "stream")
		self.stream_obj = nil
	end
	if self.sub_stream_obj then
		--Group is "pub:<id>"
		self.sub_stream_obj:Stop("pub:" .. self.id, "stream")
		self.sub_stream_obj = nil
	end

	if _M.co_service then
		_M.co_service:RemoveObject( self.expose_path )
	end

	self.mediaSource = nil
	self.mediaSource = nil
	if self.afterSwSync then
		self.afterSwSync:BindMediaSource(nil)
		self.afterSwSync = nil
	end
	if self.sub_afterSwSync then
		self.sub_afterSwSync:BindMediaSource(nil)
		self.sub_afterSwSync = nil
	end

	if self.mediaSW then
		self.mediaSW:RemoveTrack("video")
		self.mediaSW:RemoveTrack("audio")
		--TODO: Note here.
		self.mediaSW:WaitSwitchDone(100)	--Wait 100ms
		self.mediaSW = nil
	end

	if self.sub_mediaSW then
		self.sub_mediaSW:RemoveTrack("video")
		self.sub_mediaSW:RemoveTrack("audio")
		--TODO: Note here.
		self.sub_mediaSW:WaitSwitchDone(100)	--Wait 100ms
		self.sub_mediaSW = nil
	end

	collectgarbage()
end

function pubpoint_meta:Reset()
	local id = self.id
	local cfg = self.config
	self:Destroy()
	self:Init( id, cfg )
end

function pubpoint_meta:GetStreams()
	local ret = {}
	
	local sid, s
	for sid, s in pairs(self.streams) do
		if s.type ~= "ndi-hx" then
			local d = {
				id = sid,
				type = s.type,
				enable = self.config.streams and self.config.streams[sid] and self.config.streams[sid].enable or false,
				status = s.status,
				error_reason = s.error_reason or "",
				address = s.service_address or "",
				name =  self.config.streams and self.config.streams[sid] and self.config.streams[sid].name or sid
			}
			table.insert( ret,  d)
		end
	end

	table.sort( ret, function(a,b)
		return a.id < b.id
	end)

	return ret
end

function pubpoint_meta:GetStreamParameters(sid)
	if not self.config or not self.config.streams or not self.config.streams[sid] then
		return nil
	else
		if sid == 1  then
			local group_name, dev_name,ch_name =_M.GetNdiGroupName()
			self.config.streams[sid]["device_group"] = group_name
			self.config.streams[sid]["device_name"] = dev_name
			_M.full_ndiChannel = ch_name
		end
		return self.config.streams[sid]
	end
end

function pubpoint_meta:Update(special_force_id)
	local old_main = self.stream_obj
	local old_sub = self.sub_stream_obj
	if self.config.source_stream and self.config.source_stream ~= "" then
		self.stream_obj = streams.GetStreamObject( self.config.source_stream,"main" )
		self.sub_stream_obj = streams.GetStreamObject( self.config.source_stream,"sub" )
	else
		self.stream_obj = nil
		self.sub_stream_obj = nil
	end

	--Not changed.
	if (old_main == self.stream_obj and (not special_force_id or special_force_id ~= self.config.source_stream)) and (old_sub == self.sub_stream_obj and (not special_force_id or special_force_id ~= self.config.source_stream)) then
		return
	end

	local toUpdate = false

	if old_main ~= self.stream_obj then
		if self.stream_obj then
			--Group is "pub:<id>"
			--close old source_stream
			if old_main then	
				old_main:Stop("pub:" .. self.id, "stream")
			end
			self.stream_obj:Start("pub:" .. self.id, "stream")
		else
			if old_main then	
				old_main:Stop("pub:" .. self.id, "stream")
			end
		end

		if not self.mediaSW then
			if self.stream_obj then
				self.mediaSource = self.stream_obj:GetMediaSource()
			else
				self.mediaSource = nil
			end

			toUpdate = true
		else
			if self.stream_obj then
				_dbg.Info( "STREAMER", "Publish point:", self.id, " selected media from:", self.config.source_stream )
				self.mediaSW:SelectTrack( "video", self.stream_obj:GetMediaSource() )
				self.mediaSW:SelectTrack( "audio", self.stream_obj:GetMediaSource() )
				self.mediaSW:WaitSwitchDone(100)	--Wait 100ms
				
				--update media source, srt session need restart
				toUpdate = true
			else
				_dbg.Info( "STREAMER", "Publish point:", self.id, " loss the media source!" )
				self.mediaSW:SelectTrack( "video", nil, true ) --Fast switch
				self.mediaSW:SelectTrack( "audio", nil, true ) --Fast switch
				self.mediaSW:WaitSwitchDone(100)	--Wait 100ms
			end
		end
	end


	if old_sub ~= self.sub_stream_obj then
		if self.sub_stream_obj then
			--Group is "pub:<id>"
			--close old source_stream
			if old_sub then	
				old_sub:Stop("pub:" .. self.id, "stream")
			end
			self.sub_stream_obj:Start("pub:" .. self.id, "stream")
		else
			if old_sub then	
				old_sub:Stop("pub:" .. self.id, "stream")
			end
		end
		if not self.sub_mediaSW then
			if self.sub_stream_obj then
				self.sub_mediaSource = self.sub_stream_obj:GetMediaSource()
			else
				self.sub_mediaSource = nil
			end
			toUpdate = true
		else
			if self.sub_stream_obj then
				_dbg.Info( "STREAMER", "Publish point:", self.id, " selected media from sub:", self.config.source_stream )
				self.sub_mediaSW:SelectTrack( "video", self.sub_stream_obj:GetMediaSource() )
				self.sub_mediaSW:SelectTrack( "audio", self.sub_stream_obj:GetMediaSource() )
				self.sub_mediaSW:WaitSwitchDone(100)	--Wait 100ms
				
				--update media source, srt session need restart
				toUpdate = true

			else
				_dbg.Info( "STREAMER", "Publish point:", self.id, " loss the media source!" )
				self.sub_mediaSW:SelectTrack( "video", nil, true ) --Fast switch
				self.sub_mediaSW:SelectTrack( "audio", nil, true ) --Fast switch
				self.sub_mediaSW:WaitSwitchDone(100)	--Wait 100ms
			end
		end
	end
	if toUpdate then
		local id, s
		for id, s in pairs(self.streams) do
			if s.status == "started" and (s.type == "srt" or s.type == "ndi-hx") then
				self:SetStreamParameters( id, nil, true ) --Force update because srt media source changed!
			end
		end
	end
end

function pubpoint_meta:CreateStream(type,sid)
	type = type or "rtmp"
	type = type:lower()
	if type ~= "rtsp" and type ~= "rtmp" and type ~= "srt" and type ~= "udp" and type ~= "udp-ts" and type ~="ndi-hx" then
		--return false, "Only RTSP/RTMP/SRT/UDP-TS supported now"
		return false, "Err_UnSupportStreamType"
	end

	if self.streams_count >= MAX_SUPPORT_STREAMS then
		--return false, "Up to " .. MAX_SUPPORT_STREAMS .. " streams supported for a publish point"
		return false, "Err_MaxSupportStreams"
	end

	if type == "rtsp" then	--Just allow one RTSP stream for a publish point
		local x,sobj
		for x,sobj in pairs(self.streams) do
			if sobj.type == "rtsp" then
				--return false, "Only 1 RTSP service allowed in a publish point"
				return false, "Err_MaxSupportRtspService"
			end
		end
	end

	if not sid then
		sid = self.stream_id_base
		self.stream_id_base = self.stream_id_base + 1
		if not self.config.streams then
			self.config.streams = {}
		end
		self.config.streams[ sid ] = {
			type = type,
			enable = false,
			service_port = (type == "rtsp" and 554 or 0)
		}
		_M.config:__TRIGGER(true) --Save it!
	end

	local stream_obj = {
		type = type,
		status = "stopped",
		error_reason = ""
	}

	self.streams[ sid ] = stream_obj
	

	self.streams_count = self.streams_count + 1

	return sid
	-- return true
end

function pubpoint_meta:RemoveStream(sid, nosave)
	if not sid then
		return false
	end

	local stream_obj = self.streams[ sid ]
	if not stream_obj then
		return false
	end

	if stream_obj.type == "rtsp" then
		if stream_obj.rtsp_session then
			if stream_obj.rtsp_service then
				stream_obj.rtsp_service:RemoveSession( stream_obj.rtsp_session )
			end
			_dbg.Info( "STREAMER", "RTSP session:", stream_obj.rtsp_session, " is destroied for", self.expose_path, ":", sid )
			stream_obj.rtsp_session = nil
		end

		if stream_obj.rtsp_service then
			_M.DellocateRTSPServer( stream_obj.rtsp_port )
			stream_obj.rtsp_service = nil
		end
	elseif stream_obj.type == "rtmp" then
		if stream_obj.rtmp_session then
			if self.rtmp_pusher then
				self.rtmp_pusher:RemoveSession( stream_obj.rtmp_session )
			end
			_dbg.Info( "STREAMER", "RTMP:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
			stream_obj.rtmp_session = nil
			stream_obj.service_address = nil
		end
	elseif stream_obj.type == "srt" then
		if stream_obj.srt_session then
			if self.srt_pusher then
				self.srt_pusher:RemoveSession( stream_obj.srt_session )
			end
			_dbg.Info( "STREAMER", "SRT:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
			stream_obj.srt_session = nil
			stream_obj.service_address = nil
		end
	elseif stream_obj.type == "udp" or stream_obj.type == "udp-ts" then
		if stream_obj.udp_session then
			if self.net_pusher then
				self.net_pusher:RemoveSession( stream_obj.udp_session )
			end
			_dbg.Info( "STREAMER", "UDP:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
			stream_obj.udp_session = nil
			stream_obj.service_address = nil
		end
	else
		_dbg.Warn( "STREAMER", "Unsupported stream type while remove stream for", self.expose_path, ":", sid )
	end

	self.streams[ sid ] = nil
	if not nosave and self.config.streams and self.config.streams[ sid ] then
		self.config.streams[ sid ] = nil
		_M.config:__TRIGGER(true)
	end
	self.streams_count = self.streams_count - 1

	collectgarbage()
	return true
end

function pubpoint_meta:EnableStream(sid)
	return self:SetStreamParameters( sid, { enable = true } )
end

function pubpoint_meta:DisableStream(sid)
	return self:SetStreamParameters( sid, { enable = false } )
end

-------------------------------------------------
-- Get the AES Key Len.
-------------------------------------------------

function pubpoint_meta:GetAesKeyLen( aesName )
	local aesKeyLen
	if aesName == "AES-128" or aesName == "aes-128" then
		aesKeyLen = 16
	elseif aesName == "AES-192" or aesName == "aes-192"  then
		aesKeyLen = 24
	elseif aesName == "AES-256" or aesName == "aes-256" then
		aesKeyLen = 32
	else
		aesKeyLen = 0
	end
	return aesKeyLen
end

function pubpoint_meta:SetStreamParameters(sid, params, tempo,iscodec)
	if not sid then
		return false
	end


	local stream_obj = self.streams[ sid ]
	if not stream_obj then
		_dbg.Error( "STREAMER", "stream_obj is nil")
		return false
	end

	local stream_cfg = self.config.streams and self.config.streams[ sid ] or nil
	if not stream_cfg then
		_dbg.Error( "STREAMER", "stream_cfg is nil")
		return false
	end

	if sid == 1 and type(params) == "table" and not iscodec then
		if params.channel_name and _M.full_ndiChannel and  params.channel_name == _M.full_ndiChannel then
			return false,"Channel Ndi|hx has the same name as channel full Ndi"
		end
		if params.device_group and params.device_name and (params.device_group ~= stream_cfg["device_group"] or params.device_name ~= stream_cfg["device_name"]) then
			local r = _M.SetNdiGroupName(params.device_group,params.device_name,true)
			if not r then
				return false,"update ndi group error"
			end
			return true
		end
		if params.ndi_url then
			_M.ndi_url = params.ndi_url
			params.ndi_url = nil
			if self.ndi_service then
				self.ndi_service:SetWebUrl(_M.ndi_url)
			end
		end
	end

	local changes = {}
	if tempo then
		if not params then
			changes["#ANY#"] = true	--mark any properties is changed.
		else
			local k,v
			for k,v in pairs(params) do
				changes[k] = true
			end
		end
	else
		local k,v
		for k,v in pairs(params) do
			if stream_cfg[k] ~= v then
				changes[k] = true
				stream_cfg[k] = v 
	
			end
		end
		_M.config:__TRIGGER(true)	--Save.
	end




	if stream_obj.type == "rtsp" then
	------------------------------- RTSP ----------------------------
		--Session changed.
		local should_stop_session = false
		if changes["#ANY#"] or changes.service_port or changes.address or changes.sourceType then
			should_stop_session = true
		elseif changes.enable then
			should_stop_session = not stream_cfg.enable 
		end

		if should_stop_session then
			if stream_obj.rtsp_session then
				if stream_obj.rtsp_service then
					stream_obj.rtsp_service:RemoveSession( stream_obj.rtsp_session )
				end
				_dbg.Info( "STREAMER", "RTSP session:", stream_obj.rtsp_session, " is destroied for", self.expose_path, ":", sid )
				stream_obj.rtsp_session = nil
				stream_obj.service_address = nil
			end
			stream_obj.status = "stopped"
			stream_obj.error_reason = ""
		end

		--Service port changed.
		if changes["#ANY#"] or changes.service_port then
			if stream_obj.rtsp_service then
				_M.DellocateRTSPServer( stream_obj.rtsp_port )
				stream_obj.rtsp_service = nil
			end

			stream_obj.rtsp_port = stream_cfg.service_port
			if stream_obj.rtsp_port and stream_obj.rtsp_port > 0 and stream_obj.rtsp_port < 65536 then
				stream_obj.rtsp_service = _M.AllocateRTSPServer( stream_obj.rtsp_port )
				if not stream_obj.rtsp_service then
					_dbg.Error( "STREAMER", "Cannot create RTSP service (port=", stream_obj.rtsp_port, ") for", self.expose_path, ":", sid )
					stream_obj.status = "error"
					stream_obj.error_reason = "rtsp-service-fail"
				end
			else
				_dbg.Error( "STREAMER", "Invalid RTSP service port specified for", self.expose_path, ":", sid )
				stream_obj.status = "error"
				stream_obj.error_reason = "rtsp-port-invalid"
			end
		end

		--New session.
		if changes["#ANY#"] or changes.service_port or changes.address or changes.enable or changes.sourceType then
			if stream_obj.rtsp_service and stream_cfg.enable and not stream_cfg.rtsp_session then
				stream_obj.status = "starting"
				if stream_cfg.sourceType == "main" and self.mediaSource then
					if stream_obj.rtsp_service:AddSession( stream_cfg.address, self.mediaSource ) then
						_dbg.Info( "STREAMER", "Added RTSP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.rtsp_session = stream_cfg.address
						stream_obj.service_address = "rtsp://%HOST%:" .. stream_obj.rtsp_port .. "/" .. stream_cfg.address
						stream_obj.rtsp_service:Start()
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create RTSP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "rtsp-session-fail"
					end
				end

				if stream_cfg.sourceType == "sub" and self.sub_mediaSource then
					if stream_obj.rtsp_service:AddSession( stream_cfg.address, self.sub_mediaSource ) then
						_dbg.Info( "STREAMER", "Added RTSP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.rtsp_session = stream_cfg.address
						stream_obj.service_address = "rtsp://%HOST%:" .. stream_obj.rtsp_port .. "/" .. stream_cfg.address
						stream_obj.rtsp_service:Start()
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create RTSP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "rtsp-session-fail"
					end
				end
			end
		end
	----------------------------- RTSP-END --------------------------
	
	elseif stream_obj.type == "rtmp" then
	------------------------------- RTMP ----------------------------
		--Session changed.
		local should_stop_session = false
		if changes["#ANY#"] or changes.address or changes.user or changes.password or changes.sourceType then
			should_stop_session = true
		elseif changes.enable then
			should_stop_session = not stream_cfg.enable 
		end

		if should_stop_session then
			if stream_obj.rtmp_session then
				if self.rtmp_pusher then
					self.rtmp_pusher:RemoveSession( stream_obj.rtmp_session )
				end
				_dbg.Info( "STREAMER", "RTMP:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
				stream_obj.rtmp_session = nil
				stream_obj.service_address = nil
			end
			stream_obj.status = "stopped"
			stream_obj.error_reason = ""
		end

		if not self.rtmp_pusher then
			self.rtmp_pusher = KM_STREAMING.Rtmp.Pushing.createNew( "rtmp-push:" .. self.id )
			if not self.rtmp_pusher then
				_dbg.Error( "STREAMER", "Fail to create RTMP pusher for", self.expose_path, ":", sid )
				stream_obj.status = "error"
				stream_obj.error_reason = "rtmp-pusher-fail"
			else
				_dbg.Info( "STREAMER", "Created RTMP pusher for", self.expose_path, ":", sid )
			end
		end

		--New session.
		if changes["#ANY#"] or changes.user or changes.password or changes.address or changes.enable or changes.sourceType  then
			if self.rtmp_pusher and stream_cfg.enable and stream_cfg.address ~= "" and not stream_cfg.rtmp_session then
				stream_obj.status = "starting"
				if stream_cfg.sourceType == "main" and self.mediaSource then
					stream_obj.rtmp_session = self.rtmp_pusher:AddSession(
						self.mediaSource, stream_cfg.address, {
							user = stream_cfg.user, pass=stream_cfg.password,
							connTimeout = stream_cfg.timeout or 10,
							connIntv = 3, --TODO
							withMeta = true
						}
					)

					if stream_obj.rtmp_session then
						_dbg.Info( "STREAMER", "Added RTMP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = stream_cfg.address
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create RTMP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "rtmp-session-fail"
					end
				end

				if stream_cfg.sourceType == "sub" and self.sub_mediaSource then
					stream_obj.rtmp_session = self.rtmp_pusher:AddSession(
						self.sub_mediaSource, stream_cfg.address, {
							user = stream_cfg.user, pass=stream_cfg.password,
							connTimeout = stream_cfg.timeout or 10,
							connIntv = 3, --TODO
							withMeta = true
						}
					)

					if stream_obj.rtmp_session then
						_dbg.Info( "STREAMER", "Added RTMP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = stream_cfg.address
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create RTMP media session '", stream_cfg.address, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "rtmp-session-fail"
					end
				end
			end
		end
	----------------------------- RTMP-END --------------------------


	----------------------------- SRT-START ----------------------------
	elseif stream_obj.type == "srt" then

		--Session changed.
		local should_stop_session = false
		if changes["#ANY#"] or changes.address or changes.listenerPort or changes.connectionMode or changes.passphrase or changes.srt_stream_id or changes.sourceType  then
			should_stop_session = true
		elseif changes.enable then
			should_stop_session = not stream_cfg.enable 
		end

		if should_stop_session then
			if stream_obj.srt_session then
				if self.srt_pusher then
					self.srt_pusher:RemoveSession( stream_obj.srt_session )
				end
				_dbg.Info( "STREAMER", "SRT:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
				stream_obj.srt_session = nil
				stream_obj.service_address = nil
			end
			stream_obj.status = "stopped"
			stream_obj.error_reason = ""
		end

		if not self.srt_pusher then
			self.srt_pusher = KM_STREAMING.pushGroup.createNew( "srt-push:" .. self.id )
			if not self.srt_pusher then
				_dbg.Error( "STREAMER", "Fail to create SRT pusher for", self.expose_path, ":", sid )
				stream_obj.status = "error"
				stream_obj.error_reason = "srt-pusher-fail"
			else
				_dbg.Info( "STREAMER", "Created SRT pusher for", self.expose_path, ":", sid )
			end
		end
		--_dbg.Error( "STREAMER", " Create SRT pusher for listener enable:" ,changes.enable, " address:",changes.address," listenerPort:",changes.listenerPort)
		--New session.
		if changes["#ANY#"] or changes.connectionMode or changes.passphrase or changes.address or changes.listenerPort or changes.enable or changes.srt_stream_id or changes.sourceType then
			--modify by yj@2020-01-14 srt client support listener mode,but address is null
			--if self.srt_pusher and stream_cfg.enable and stream_cfg.address ~= "" and not stream_cfg.srt_session then
			if self.srt_pusher and (not stream_obj.srt_session) and stream_cfg.enable and (stream_cfg.connectionMode and stream_cfg.connectionMode ~= "") 
			and (stream_cfg.listenerPort and stream_cfg.listenerPort ~= "") and (stream_cfg.payloadSize and stream_cfg.payloadSize  ~= "") then
				stream_obj.status = "starting"
				--check srt para
				local srtAddress
				if stream_cfg.address == "undefined" then
					srtAddress = ""
				else
					srtAddress = stream_cfg.address
				end
				if stream_cfg.connectionMode == "listener" or stream_cfg.connectionMode == "Listener" then
					srtAddress = ""
				end
				local srtUrl = "srt://" .. srtAddress .. ":" .. stream_cfg.listenerPort
				
				if stream_cfg.sourceType=="main" and self.mediaSource then
					stream_obj.srt_session = self.srt_pusher:AddSession("stream"..sid, self.mediaSource, {videoBuffer=8*1024*1024,audioBuffer=64*1024})
					if stream_obj.srt_session then
						self.srt_pusher:SetOption(stream_obj.srt_session,{PushPath = srtUrl,PushMode = "NONE",PushProtocolType = "srt",
																		srt_connect_mode = stream_cfg.connectionMode,
																		srt_encryption_key_len =  self:GetAesKeyLen(stream_cfg.encryption),
																		srt_encryption_key = stream_cfg.passphrase,
																		srt_latency = stream_cfg.latency, 
																		srt_bandwidth_overhead = stream_cfg.bandwidth * 0.01, 
																		srt_payload_size = stream_cfg.payloadSize,
																		srt_stream_id = stream_cfg.srt_stream_id or ""})
						self.srt_pusher:StartSession(stream_obj.srt_session)
						_dbg.Info( "STREAMER", "Added SRT media session '", srtUrl, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = srtUrl
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create SRT media session '", srtUrl, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "srt_session-fail"
					end
				end

				if stream_cfg.sourceType == "sub" and self.sub_mediaSource then
					stream_obj.srt_session = self.srt_pusher:AddSession("stream"..sid, self.sub_mediaSource, {videoBuffer=8*1024*1024,audioBuffer=64*1024})
					if stream_obj.srt_session then
						self.srt_pusher:SetOption(stream_obj.srt_session,{PushPath = srtUrl,PushMode = "NONE",PushProtocolType = "srt",
																		srt_connect_mode = stream_cfg.connectionMode,
																		srt_encryption_key_len =  self:GetAesKeyLen(stream_cfg.encryption),
																		srt_encryption_key = stream_cfg.passphrase,
																		srt_latency = stream_cfg.latency, 
																		srt_bandwidth_overhead = stream_cfg.bandwidth * 0.01, 
																		srt_payload_size = stream_cfg.payloadSize,
																		srt_stream_id = stream_cfg.srt_stream_id or ""})
						self.srt_pusher:StartSession(stream_obj.srt_session)
						_dbg.Info( "STREAMER", "Added SRT media session '", srtUrl, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = srtUrl
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create SRT media session '", srtUrl, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "srt_session-fail"
					end
				end
			end
		end

		-- self.mediaSource, stream_cfg.address, {
		-- 	user = stream_cfg.user, pass=stream_cfg.password,
		-- 	connTimeout = stream_cfg.timeout or 10,
		-- 	connIntv = 3, --TODO
		-- 	withMeta = true
		-- }


	----------------------------- SRT-END ------------------------------

	----------------------------- UDP-TS-START ----------------------------
	elseif stream_obj.type == "udp" or stream_obj.type == "udp-ts" then
		local should_stop_session = false
		if changes["#ANY#"] or changes.address or changes.port  or changes.ts_transport_stream_id or changes.ts_null_multiple
		or changes.ts_pcr_period  or changes.ts_pts_pcr_delay or changes.ts_service_name or changes.sourceType then
			should_stop_session = true
		elseif changes.enable then
			should_stop_session = not stream_cfg.enable
		end

		if should_stop_session then
			if stream_obj.udp_session then
				if self.net_pusher then
					self.net_pusher:RemoveSession( stream_obj.udp_session )
				end
				_dbg.Info( "STREAMER", "UDP-TS:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
				stream_obj.udp_session = nil
				stream_obj.service_address = nil
			end
			stream_obj.status = "stopped"
			stream_obj.error_reason = ""
		end

		if not self.net_pusher then
			self.net_pusher = KM_STREAMING.NetPusher.createNew( "udp-push:" .. self.id )
			if not self.net_pusher then
				_dbg.Error( "STREAMER", "Fail to create udp pusher for", self.expose_path, ":", sid )
				stream_obj.status = "error"
				stream_obj.error_reason = "udp-pusher-fail"
			else
				_dbg.Info( "STREAMER", "Created udp pusher for", self.expose_path, ":", sid )
			end
		end

		--New session.
		if changes["#ANY#"] or changes.address or changes.port or changes.enable or changes.ts_transport_stream_id or changes.ts_null_multiple
		or changes.ts_pcr_period  or changes.ts_pts_pcr_delay or changes.ts_service_name or changes.sourceType then
			--check params
			if self.net_pusher and (not stream_obj.udp_session) and stream_cfg.enable and (stream_cfg.address and stream_cfg.address ~= "")
			    and (stream_cfg.port and stream_cfg.port ~= "") then
				stream_obj.status = "starting"

				if not stream_cfg.ttl then
					stream_cfg.ttl = "127"
				end
				-- How to get muxrate ????
				local muxrate = -1
				local ts_null_multiple = "2.0"
				if stream_cfg.ts_null_multiple == "0" or stream_cfg.ts_null_multiple == "" then
					muxrate = -1
				    ts_null_multiple = "1.0"
				else
					muxrate = stream_cfg.ts_null_multiple / 2.0
				end
				--muxrate = -1
				local targetAddr = "udp://" .. stream_cfg.address .. ":" .. stream_cfg.port
				local pushPath = targetAddr .."?" .."ttl=".. stream_cfg.ttl .."&pkt_size=1316"
				
				if stream_cfg.sourceType == "main" and self.mediaSource then
					stream_obj.udp_session = self.net_pusher:AddSession("udp-stream"..sid, self.mediaSource, {videoBuffer=8*1024*1024,audioBuffer=64*1024})
					if stream_obj.udp_session then
						self.net_pusher:SetOption(stream_obj.udp_session,
												{
													PushPath = pushPath,
													PushMode = "NONE",
													PushProtocolType = "udp",
													muxrate  = muxrate,
													ts_transport_stream_id =  stream_cfg.ts_transport_stream_id,
													ts_pts_pcr_delay  = stream_cfg.ts_pts_pcr_delay,
													ts_pmt_start_pid  = stream_cfg.ts_pmt_start_pid,
													ts_start_pid  	  = stream_cfg.ts_start_pid,
													ts_tables_version = stream_cfg.ts_tables_version,
													ts_service_name   =  stream_cfg.ts_service_name,
													ts_service_provider	=  stream_cfg.ts_service_provider ,
													ts_null_multiple    =  ts_null_multiple ,
													ts_pcr_period       =  stream_cfg.ts_pcr_period
												})
						self.net_pusher:StartSession(stream_obj.udp_session)
						_dbg.Info( "STREAMER", "Added udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = targetAddr
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "udp_session-fail"
					end
				end

				if stream_cfg.sourceType == "sub" and self.sub_mediaSource then
					stream_obj.udp_session = self.net_pusher:AddSession("udp-stream"..sid, self.sub_mediaSource, {videoBuffer=8*1024*1024,audioBuffer=64*1024})
					if stream_obj.udp_session then
						self.net_pusher:SetOption(stream_obj.udp_session,
												{
													PushPath = pushPath,
													PushMode = "NONE",
													PushProtocolType = "udp",
													muxrate  = muxrate,
													ts_transport_stream_id =  stream_cfg.ts_transport_stream_id,
													ts_pts_pcr_delay  = stream_cfg.ts_pts_pcr_delay,
													ts_pmt_start_pid  = stream_cfg.ts_pmt_start_pid,
													ts_start_pid  	  = stream_cfg.ts_start_pid,
													ts_tables_version = stream_cfg.ts_tables_version,
													ts_service_name   =  stream_cfg.ts_service_name,
													ts_service_provider	=  stream_cfg.ts_service_provider ,
													ts_null_multiple    =  ts_null_multiple ,
													ts_pcr_period       =  stream_cfg.ts_pcr_period
												})
						self.net_pusher:StartSession(stream_obj.udp_session)
						_dbg.Info( "STREAMER", "Added udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
						stream_obj.service_address = targetAddr
						stream_obj.status = "started"
					else
						_dbg.Error( "STREAMER", "Fail to create udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
						stream_obj.status = "error"
						stream_obj.error_reason = "udp_session-fail"
					end
				end
			end
		end
----------------------------- UDP-TS-END ------------------------------

------------------------------ NDI-HX ---------------------------------
	elseif stream_obj.type == "ndi" or stream_obj.type == "ndi-hx" or stream_obj.type == "hx" then
		local should_stop_session = false
		if changes["#ANY#"] or changes.channel_name or changes.device_group or changes.device_name  or changes.netmask or changes.ndi_connection
		or changes.netprefix  or changes.server_addr or changes.ttl then
			should_stop_session = true
		elseif changes.enable then
			should_stop_session = not stream_cfg.enable
		end

		if should_stop_session then -- 有改变，之前未启动
			if stream_obj.ndi_session then
				if self.ndi_service then
					-- self.ndi_service:StopAllSessions()
					self.ndi_service:RemoveAllSessions()
				end
				_dbg.Info( "STREAMER", "NDI-HX:", stream_obj.service_address, " is destroied for", self.expose_path, ":", sid )
				stream_obj.ndi_session = nil
				stream_obj.service_address = nil

			end
		stream_obj.status = "stopped"
		stream_obj.error_reason = ""
		end

		if not self.ndi_service then
			self.ndi_service = KM_STREAMING.ndiMediaSender.createNew()
			if not self.ndi_service then
				_dbg.Error( "STREAMER", "Fail to create NDI-HX pusher for", self.expose_path, ":", sid )
				stream_obj.status = "error"
				stream_obj.error_reason = "ndi-hx-pusher-fail"
				return true
			else
				_dbg.Info( "STREAMER", "Created ndi-hx pusher for", self.expose_path, ":", sid )
			end
		end

		--New session.
		if changes["#ANY#"] or changes.channel_name or changes.device_group or changes.device_name  or changes.netmask or changes.ndi_connection
		or changes.netprefix  or changes.server_addr or changes.ttl or changes.enable then

			_dbg.Info( "STREAMER", "~~~~~~~~~~~~~~~check params")

			-- local pt = require "pl.pretty"
			-- pt.dump(stream_cfg)

			--check params
			if self.ndi_service and (not stream_obj.ndi_session) and stream_cfg.enable and (stream_cfg.channel_name and stream_cfg.channel_name ~= "")
				and (stream_cfg.ndi_connection and stream_cfg.ndi_connection ~= "") then
				stream_obj.status = "starting"
				if stream_obj.ndi_session then
					-- self.ndi_service:StopAllSessions()
					self.ndi_service:RemoveAllSessions()
				end

				local group = string.gsub(stream_cfg.device_group," ","")

				local targetAddr = "ndi://" .. stream_cfg.channel_name
				self.ndi_service:SetOption({
					ttl = tostring(stream_cfg.ttl),
					name = stream_cfg.channel_name,
					group = group,
					netmask = stream_cfg.netmask,
					connection = stream_cfg.ndi_connection,
					netprefix = stream_cfg.netprefix,
					discovery_server = stream_cfg.server_addr ~= "" and stream_cfg.server_addr or nil
				})

				_dbg.Info( "STREAMER", "~~~~~~~~~~~~~~~RemoveAllSessions")

				if not self.mediaSource then
					_dbg.Error( "STREAMER", "Fail to create udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
					stream_obj.status = "error"
					stream_obj.error_reason = "udp_session-fail"
					stream_obj.ndi_session = nil
					return true
				end

				if self.mediaSource and self.stream_obj then
					_dbg.Info( "STREAMER", "Added udp-ts media session '", pushPath, "' for", self.expose_path, ":", sid )
					self.ndi_service:AddSession(0, self.mediaSource, {videoBuffer = 8*1024*1024, audioBuffer = 64*1024})
					if self.sub_mediaSource and self.sub_stream_obj then
						self.ndi_service:AddSession(1, self.sub_mediaSource, {videoBuffer = 8*1024*1024, audioBuffer = 64*1024})
					end
				end
				stream_obj.ndi_session = true
				stream_obj.service_address = targetAddr
				stream_obj.status = "started"   
				self.ndi_service:SetWebUrl(_M.ndi_url)
				self.ndi_service:StartAllSessions()

			end
		end

	------------------------------ NDI-HX END ---------------------------------	
	
	else
		_dbg.Warn( "STREAMER", "Unsupported stream type while set stream parameters for", self.expose_path, ":", sid )
	end
	return true
end

function pubpoint_meta:SelectSource( source_stream_id )
	self.config.source_stream = source_stream_id
	_M.config:__TRIGGER(true)
	self:Update()
	return true
end



pubpoint_meta.__index = pubpoint_meta

-------------------------------------

local function newPublishPoint()
	return setmetatable({}, pubpoint_meta)
end


function _M.SetNdiGroupName(group,name)
	local proxy = CO.GetProxy( "/", "codec" )
	if not proxy then
		_dbg.Error("STREAMER", "CoManager codec connect error" )
		return false
	else
		local r,Data = proxy.SetNameWithGroup(group,name)
		proxy:Destroy()
		if not r or not Data then
			_dbg.Error("STREAMER", "SetNdiGroupName error" )
			return false
		else
			return true
		end
	end
	return false
end

function _M.GetNdiGroupName()
	local proxy = CO.GetProxy( "/", "codec" )
	if not proxy then
		_dbg.Error("STREAMER", "CoManager codec connect error" )
		return false
	else
		local r, group_name, dev_name, chn_name = proxy.GetNameWithGroup()
		proxy:Destroy()
		if not r then
			_dbg.Error("STREAMER", "GetNdiGroupName error" )
			return false
		else
			return group_name,dev_name,chn_name
		end
	end
	return false
end

function _M.UpdateNdiGroupName(args)
	if args and _M.pub_points["1"] then
		_M.pub_points["1"]:SetStreamParameters(1,args,false,true)
		return true
	end
	return false
end


function _M.Init(cosvc)
	_M.co_service = cosvc
	if not cosvc then
		_dbg.Warn( "STREAMER", "Invalid Co-Service specified for streamer initialization, that may work not correctly!" )
	end

	_M.config = cfg_streamer.Load()
	if not _M.config then
		_dbg.Error( "STREAMER", "Fail to load streamer configure!" )
		_M.config = {
			__TRIGGER = function() end
		}
	end

	if features.FEATURE_LOCAL_STREAM then
		local modulenum = 4
		--[[
		local mo_sn = require "system.serialNo"
	
		if (mo_sn) then
			_M.serialno = mo_sn.GetSerialNo()
			local devtype = string.sub(_M.serialno,3,4)
			if (devtype == "E4") then
				modulenum = 4
			elseif (devtype == "E2") then
				modulenum = 2
			else
				modulenum = 4
			end
		end
		--]]

		local k,localcfg,localsubcfg
		for k=1, modulenum do
			local streamername
			if(k == 1) then
				streamername="HDMI"
			elseif (k == 2) then
				streamername="SDI-1"
			elseif (k == 3) then
				streamername="SDI-2"
			elseif (k == 4) then
				streamername="SDI-3"
			end

			-- localcfg = {
			-- 	source_stream = "LOCAL-0" .. (k-1),
	        --     name = streamername,
			-- 	streams = {
			-- 		{
			-- 			type = "rtsp",
			-- 			id = "1",
			-- 			enable = true,
			-- 			service_port = 554,
			-- 			address = "LOCAL-0" .. (k-1) .. ".coaf"
			-- 		}
			-- 	}
			-- }
			-- local point = newPublishPoint()
			-- if not point then
			-- 	_dbg.Warn( "STREAMER", "Fail to create publish point for local streams!")
			-- else
			-- 	_M.pub_points[ tostring(k) ] = point
			-- 	if not point then
			-- 		_dbg.Warn( "STREAMER", "Fail to create publish point for ID:", k, ", name:", localcfg.name )
			-- 	else
			-- 		point:Init( tostring(k), localcfg )
			-- 		_M.pub_points_count = _M.pub_points_count + 1
			-- 	end
			-- end

			localcfg = {
				source_stream = "LOCAL-0" .. (k-1),
	            name = streamername,
				streams = {
					{
						type = "rtsp",
						id = "1",
						enable = true,
						service_port = 554,
						address = "LOCAL-0" .. (k-1) .. ".coaf"
					}
				}
			}
			local point = newPublishPoint()
			if not point then
				_dbg.Warn( "STREAMER", "Fail to create publish point for local streams!")
			else
				_M.pub_points[ tostring(2*k-1) ] = point
				if not point then
					_dbg.Warn( "STREAMER", "Fail to create publish point for ID:", 2*k-1, ", name:", localcfg.name )
				else
					point:Init( tostring(2*k-1), localcfg )
					_M.pub_points_count = _M.pub_points_count + 1
				end
			end

			localsubcfg = {
				source_stream = "LOCAL-0" .. (k-1) .. ":sub",
	            name = streamername..":sub",
				streams = {
					{
						type = "rtsp",
						id = "2",
						enable = true,
						service_port = 554,
						address = "LOCAL-0" .. (k-1) .. "sub.coaf"
					}
				}
			}


			local subpoint = newPublishPoint()
			if not subpoint then
				_dbg.Warn( "STREAMER", "Fail to create publish sub point for local sub streams!")
			else
				_M.pub_points[ tostring(k*2) ] = subpoint
				if not subpoint then
					_dbg.Warn( "STREAMER", "Fail to create publish sub point for ID:", k*2, ", name:", localsubcfg.name )
				else
					subpoint:Init( tostring(k*2), localsubcfg )
					_M.pub_points_count = _M.pub_points_count + 1
				end
			end

		end
	end


	local id, cfg
	local max_id
	for id, cfg in pairs(_M.config) do
		if not max_id or tonumber(id) > max_id then
			max_id = tonumber(id)
		end
		local point = newPublishPoint()
		if not point then
			_dbg.Warn( "STREAMER", "Fail to create publish point for ID:", id, ", name:", cfg.name )
		else
			_M.pub_points[ id ] = point
			point:Init( id, cfg )
			_M.pub_points_count = _M.pub_points_count + 1
		end
	end

	if max_id then
		_M.pubpoint_id_base = max_id + 1
	end
end

function _M.AllocateRTSPServer( port, httpTunelPort )
	if _M.rtsp_servers[port] then
		_M.rtsp_servers[port].ref_counter = _M.rtsp_servers[port].ref_counter + 1
		return _M.rtsp_servers[port].server
	else
		local rtspServer = KM_STREAMING.Rtsp.Server.createNew( 'KILOVIEW Rtsp Media Server', "*", port, httpTunelPort )
		if not rtspServer then
			_dbg.Error( "STREAMER", "Fail to create RTSP server of port:", port, ", http-tunel-port:", httpTunelPort )
			return nil
		else
			_dbg.Info( "STREAMER", "RTSP server created, with port:", port, ", http-tunel-port:", httpTunelPort )
		end

		_M.rtsp_servers[port] = {
			ref_counter = 1,
			server = rtspServer
		}

		return rtspServer
	end
end

function _M.DellocateRTSPServer( port )
	if _M.rtsp_servers[port] then
		if _M.rtsp_servers[port].ref_counter > 0 then
			_M.rtsp_servers[port].ref_counter = _M.rtsp_servers[port].ref_counter - 1
		end
		--So, when on some check point, if found 'ref_counter' <= 0, drop the RTSP server instance.
	end
end

function _M.GetPublishPoints(spec_id)
	local ret = {}
	local id, point
	for id, point in pairs(_M.pub_points) do
		if not spec_id and tonumber(id) > 0 then
			local streamername
			if features.FEATURE_LOCAL_STREAM then
				if(tonumber(id) < 9) then
					-- if(id == '1') then
					-- 	streamername="HDMI"
					-- elseif (id == '2') then
					-- 	streamername="SDI-1"
					-- elseif (id == '3') then
					-- 	streamername="SDI-2"
					-- elseif (id == '4') then
					-- 	streamername="SDI-3"
					-- end
					if(id == '1') then
						streamername="HDMI"
					elseif (id == '2') then
						streamername="HDMI:sub"
					elseif (id == '3') then
						streamername="SDI-1"
					elseif (id == '4') then
						streamername="SDI-1:sub"
					elseif (id == '5') then
						streamername="SDI-2"
					elseif (id == '6') then
						streamername="SDI-2:sub"
					elseif (id == '7') then
						streamername="SDI-3"
					elseif (id == '8') then
						streamername="SDI-3:sub"
					end
				else
					streamername = _M.config[ id ] and _M.config[ id ].name or "UNKNOWN"
				end
			else --D300
				streamername = _M.config[ id ] and _M.config[ id ].name or "UNKNOWN"
			end

			table.insert( ret, {
				id = point.id,
				objPath = point.expose_path,
				name = streamername
			})
		elseif spec_id == id then
			return {
				id = point.id,
				objPath = point.expose_path,
				name = _M.config[ id ] and _M.config[ id ].name or "UNKNOWN"
			}
		end
	end

	if spec_id then
		return nil
	end

	table.sort( ret, function(a,b)
		return a.id < b.id
	end )
	return ret
end

function _M.GetPublishPointStatus(point_id)
	if not point_id or not _M.pub_points[point_id] then
		return {
			streams = 0,
			bitrate = 0,
			total_br = "0",
			current_stream = "",
			current_stream_name = ""
		}
	end

	local point = _M.pub_points[point_id]

	local detail = streams.GetStreamDetail( point.config.source_stream ,"main" )

	return {
		streams = point.streams_count,
		bitrate = detail and detail.stream_info and detail.stream_info.bitrate or 0,
		total_br = "unknown", --TODO
		current_stream = point.config.source_stream or "",
		current_stream_name = detail and detail.name or "none"
	}
end

function _M.ResetPublishPoint(point_id)
	if not point_id or not _M.pub_points[point_id] then
		return false
	end

	local point = _M.pub_points[point_id]
	point:Reset()
	return true
end

function _M.AddPublishPoint(name)
	if not name then
		return false, "Invalid publish point name"
	end
	if _M.pub_points_count >= MAX_SUPPORT_PUBPOINTS then
		return false, "Max " .. MAX_SUPPORT_PUBPOINTS .. " publish points allowed"
	end

	local point = newPublishPoint()
	if not point then
		_dbg.Error( "STREAMER", "Fail to create publish point while adding it, name:", name )
		return false, "Fail create publish point"
	end

	local point_id = _M.pubpoint_id_base
	_M.pubpoint_id_base = _M.pubpoint_id_base + 1
	point_id = tostring( point_id )

	_M.config[ point_id ] = {
		name = name,
		source_stream = "",
		streams = {}
	}

	_M.config:__TRIGGER(true)

	_M.pub_points[ point_id ] = point
	point:Init( point_id, _M.config[point_id] )
	_M.pub_points_count = _M.pub_points_count + 1

	return true
end

function _M.RenamePublishPoint(point_id,name)
	if not point_id then
		return false
	end
	local cfg = _M.config[ point_id ]
	if not cfg then
		return false
	end
	
	cfg.name = name or cfg.name
	_M.config:__TRIGGER(true)

	return true
end

function _M.RemovePublishPoint(point_id)
	if not point_id then
		return false
	end

	local point = _M.pub_points[ point_id ]
	if not point then
		return false
	end

	point:Destroy()
	_M.pub_points[ point_id ] = nil
	if _M.pub_points_count > 0 then
		_M.pub_points_count = _M.pub_points_count - 1
	end

	_M.config[ point_id ] = nil
	_M.config:__TRIGGER(true)

	collectgarbage()
	return true
end

function _M.UpdateStreams(special_force_id)
	local id, point
	for id, point in pairs(_M.pub_points) do
		point:Update(special_force_id)
	end
end


function _M.Step()
	local proxy = CO.GetProxy( "/", "codec" )
	if not proxy then
		_dbg.Error("STREAMER", "CoManager codec connect error" )
	else
		local r,url = proxy.GetNdiUrl()
		proxy:Destroy()
		if not r or not url then
			_dbg.Error("STREAMER", "SetNdiGroupName error" )
		else
			if _M.ndi_url ~= url then
				_M.ndi_url = url
				if _M.pub_points["1"] then
					_M.pub_points["1"]:SetStreamParameters(1,{ndi_url = _M.ndi_url})
				end
			end
		end
	end
end

function _M.IsBlocked()
	return true
end

return _M

